HEALTH = {
    "big rock": 45,
    "bonsai pot": 24,
    "blue dye": 6,
    "bluebell": 10,
    "blueprint": 8,
    "bowl": 6,
    "brick": 12,
    "campfire": 8,
    "carroot": 8,
    "carrot": 4,
    "clay": 12,
    "coal ore": 20,
    "coal": 8,
    "composter": 8,
    "copper brick": 12,
    "copper brick floor": 8,
    "corpse": 1,
    "dirt": 5,
    "fertilizer": 5,
    "flint axe": 6,
    "flint hammer": 10,
    "flint pickaxe": 6,
    "flint shovel": 6,
    "flint sword": 6,
    "flint": 16,
    "glass": 8,
    "glass key": 1,
    "glass lock": 1,
    "ice": 12,
    "junk": 1,
    "left": 1,
    "log floor": 12,
    "log": 8,
    "manual press": 12,
    "masonry bench": 12,
    "mushroom block": 5,
    "mushroom door": 10,
    "mushroom floor": 8,
    "mushroom shaper": 13,
    "mushroom stew": 6,
    "mushroom": 10,
    "obelisk": 1,
    "paper": 8,
    "pebble": 8,
    "plant bouquet": 6,
    "player": 20,
    "rabbit adult": 10,
    "rabbit child": 6,
    "rabbit fur": 8,
    "rabbit hole": 1,
    "rabbit meat": 6,
    "roasted mushroom": 8,
    "roasted rabbit meat": 10,
    "rock": 20,
    "rock axe": 10,
    "rock pickaxe": 10,
    "rock shovel": 10,
    "rock sword": 10,
    "sand": 6,
    "sapling": 4,
    "sawbench": 12,
    "small barrel": 10,
    "small crate": 10,
    "spore": 4,
    "stick": 4,
    "stone brick": 16,
    "stone": 16,
    "tree": 10,
    "treeling": 9,
    "up": 1,
    "void": 1,
    "water": 8,
    "wood door": 8,
    "wood floor": 6,
    "wood pulp": 6,
    "wood": 4,
    "wooden bed": 12,
    "wooden table": 8,
    "workbench": 5,
}


from math import cos, pi

import pygame as pg

from ...info import TILE_SIZE, HALF_SIZE, CHUNK_SIZE, SCREEN_SIZE, DAY_LENGTH, ATTRIBUTES

def create_light_surface(intensity: int, color: tuple[int, int, int]):
    light_surface = pg.Surface((32, 32), pg.SRCALPHA)
    for i in range(16, 0, -1):
        alpha = intensity - int((i / 16) * intensity)
        pg.draw.circle(light_surface, (*color, alpha), (16, 16), i)
    return light_surface

LIGHTS = {"campfire": (create_light_surface(170, (181, 102, 60)), 832)}

def render_lights(tick, chunks, location, zoom, camera, window):
    dark_overlay = pg.Surface(SCREEN_SIZE)
    dark_overlay.fill((19, 17, 18))
    dark_overlay.set_alpha(int((1 - cos(((tick / DAY_LENGTH * 2) - 1 / 2) * pi)) * 95))
    window.blit(dark_overlay, (0, 0))

    for x in range(-3, 4):
        for y in range(-3, 4):
            chunk = (x + location["tile"][0], y + location["tile"][1])
            if chunk in chunks:
                for tile in chunks[chunk]:
                    current_tile = chunks[chunk][tile]
                    if "kind" in current_tile and "light" in ATTRIBUTES.get(current_tile["kind"], ()):
                        scaled_glow = pg.transform.scale(LIGHTS[current_tile["kind"]][0], (int(LIGHTS[current_tile["kind"]][1] * zoom), int(LIGHTS[current_tile["kind"]][1] * zoom)))
                        night_factor = 1 - cos(((tick / DAY_LENGTH * 2) - 1 / 2) * pi)
                        scaled_glow.set_alpha(int(night_factor * 180))
                        placement_x = camera[0] + (tile[0] * TILE_SIZE + chunk[0] * CHUNK_SIZE + HALF_SIZE) * zoom - int(LIGHTS[current_tile["kind"]][1] * zoom / 2)
                        placement_y = camera[1] + (tile[1] * TILE_SIZE + chunk[1] * CHUNK_SIZE + HALF_SIZE) * zoom - int(LIGHTS[current_tile["kind"]][1] * zoom / 2)
                        window.blit(scaled_glow, (placement_x, placement_y))
    return window


def left(chunks, chunk, tile, delete_tiles):
    left_chunk = (chunk[0] + (tile[0] - 1) // 16, chunk[1])
    left_tile = ((tile[0] - 1) % 16, tile[1])
    if left_tile not in chunks.get(left_chunk, {}):
        delete_tiles.append((chunk, tile))
    elif "kind" not in chunks[left_chunk][left_tile]:
        del chunks[chunk][tile]["kind"]
    return chunks, delete_tiles

def up(chunks, chunk, tile, delete_tiles):
    up_chunk = (chunk[0], chunk[1] + (tile[1] - 1) // 16)
    up_tile = (tile[0], (tile[1] - 1) % 16)
    if up_tile not in chunks.get(up_chunk, {}):
        delete_tiles.append((chunk, tile))
    elif "kind" not in chunks[up_chunk][up_tile]:
        del chunks[chunk][tile]["kind"]
    return chunks, delete_tiles


from ..info import ATTRIBUTES, MULTI_TILES

def place_tile(kind, grid_position, chunks):
    if "multi" in ATTRIBUTES.get(kind, ()):
        width, height = MULTI_TILES[kind]
        for x in range(width):
            for y in range(height):
                chunk_pos = (grid_position[0][0] + (grid_position[1][0] + x) // 16, grid_position[0][1] + (grid_position[1][1] + y) // 16)
                tile_pos = ((grid_position[1][0] + x) % 16, (grid_position[1][1] + y) % 16)
                tile_type = "left" if y == 0 else "up"
                old_tile = chunks[chunk_pos].get(tile_pos)
                if old_tile:
                    chunks[chunk_pos][tile_pos] = {"kind": tile_type, "floor": old_tile["floor"]}
                else:
                    chunks[chunk_pos][tile_pos] = {"kind": tile_type}
    if grid_position[1] not in chunks[grid_position[0]]:
        chunks[grid_position[0]][grid_position[1]] = {"kind": kind}
    else:
        old_tile = chunks[grid_position[0]][grid_position[1]]
        chunks[grid_position[0]][grid_position[1]] = {"kind": kind, "floor": old_tile["floor"]}
    return chunks


from ...world_generation.world_generation import generate_chunk

def save_creating(state, chunks):
    state.save_file_name = ""
    state.menu_placement = "main_game"
    chunks = {}
    state.location["tile"] = [0, 0, 0, 2]
    state.location["real"] = [0, 0, 0, 2]
    state.noise_offset = generate_chunk(state.world_type, 0, 0, chunks)
    for x in range(-4, 5):
        for y in range(-4, 5):
            generate_chunk(state.world_type, state.location["tile"][0] + x, state.location["tile"][1] + y, chunks, state.noise_offset)
    chunks[0, 0][0, 0] = {"kind": "obelisk", "health": 1}
    chunks[0, 0][0, 1] = {"kind": "up", "health": 1}
    chunks[0, 0][0, 2] = {"kind": "player", "inventory": {}, "health": 20, "floor": "void", "recipe": 0}
    if state.world_type == 1:
        chunks[0, 0][0, 2]["inventory"] = {"flint axe": 1}
        chunks[0, 0][0, 3] = {"kind": "tree", "floor": "dirt", "inventory": {"log": 2, "sapling": 2}}
        chunks[0, 0][0, 4] = {"kind": "composter"}
    state.tick = 0
    return chunks


from ...info import SCREEN_SIZE, INVENTORY_SIZE, UI_SCALE, STORAGE
from .put_in import put_in
from .take_out import take_out

def open_storage(position, chunks, location, inventory, machine_ui):
    moved_x = position[0] - SCREEN_SIZE[0] // 2
    if "inventory" not in chunks[location["opened"][0]][location["opened"][1]]:
        chunks[location["opened"][0]][location["opened"][1]]["inventory"] = {}
    machine_inventory = chunks[location["opened"][0]][location["opened"][1]]["inventory"]
    holding_over_inventory = position[1] >= SCREEN_SIZE[1] - 32 * UI_SCALE and abs(moved_x) <= 16 * INVENTORY_SIZE[0] * UI_SCALE
    if holding_over_inventory:
        slot_number = ((moved_x - 16 * UI_SCALE * (INVENTORY_SIZE[0] % 2)) // (32 * UI_SCALE) + INVENTORY_SIZE[0] // 2 + INVENTORY_SIZE[0] % 2)
        machine_storage = STORAGE.get(machine_ui, (14, 16))
        chunks = put_in(chunks, location, inventory, machine_storage, slot_number, machine_inventory)
    elif (SCREEN_SIZE[1] - 144 * UI_SCALE <= position[1] <= SCREEN_SIZE[1] - 80 * UI_SCALE and abs(moved_x) <= 112 * UI_SCALE):
        slot_number = (moved_x + 112 * UI_SCALE) // (32 * UI_SCALE) + (position[1] - SCREEN_SIZE[1] + 144 * UI_SCALE) // (32 * UI_SCALE) * 7
        if slot_number < len(machine_inventory):
            item = list(machine_inventory.items())[slot_number]
            chunks = take_out(chunks, location, inventory, item)
    return chunks

def closed_storage(chunks, grid_position, inventory, location, inventory_number):
    location["opened"] = (grid_position[0], grid_position[1])
    if "inventory" in chunks[grid_position[0]][grid_position[1]]:
        take_out(chunks, location, inventory, list(chunks[grid_position[0]][grid_position[1]]["inventory"].items())[0])
    elif inventory_number < len(inventory):
        chunks[grid_position[0]][grid_position[1]]["inventory"] = {}
        put_in(chunks, location, inventory, (1, 64), inventory_number, {})
    return chunks


from ...info import FOOD, FLOOR, ATTRIBUTES, HEALTH
from ...other_systems.tile_placement import place_tile
from ...other_systems.tile_placable import is_placable


def place(inventory, inventory_number, is_not_tile, is_kind, health, grid_position, location, chunks):
    if len(inventory) > inventory_number:
        inventory_key = list(inventory.keys())[inventory_number]
        if inventory_key not in FLOOR:
            if is_not_tile or not is_kind:
                if "eat" in ATTRIBUTES.get(inventory_key, ()):
                    if health < HEALTH["player"]:
                        chunks[location["tile"][0], location["tile"][1]][location["tile"][2], location["tile"][3]]["health"] = min(health + FOOD[inventory_key], HEALTH["player"])
                        inventory[inventory_key] -= 1
                        if inventory[inventory_key] == 0:
                            del inventory[inventory_key]
                        return chunks
                if is_placable(inventory_key, grid_position, chunks):
                    chunks = place_tile(inventory_key, grid_position, chunks)
                    inventory[inventory_key] -= 1
        elif is_not_tile:
            inventory[inventory_key] -= 1
            chunks[grid_position[0]][grid_position[1]] = {"floor": inventory_key}
        if inventory[inventory_key] == 0:
            del inventory[inventory_key]
    return chunks


from ..right_click import break_tile, break_floor
from ...info import UNBREAK

def right_click(
    chunks,
    grid_position: tuple[tuple[int, int], tuple[int, int]],
    inventory: dict[str, int],
    inventory_number: int,
    location: dict,
    machine_ui: str,
):
    machine_ui = "game"
    if grid_position[1] not in chunks[grid_position[0]]:
        return chunks, location, machine_ui
    mining_tile = chunks[grid_position[0]][grid_position[1]]
    delete_mining_tile = False
    player_tile = chunks[location["tile"][0], location["tile"][1]][location["tile"][2], location["tile"][3]]
    location["mined"] = (grid_position[0], grid_position[1])
    if "kind" in mining_tile and mining_tile["kind"] not in UNBREAK:
        chunks, delete_mining_tile, mining_tile = break_tile(mining_tile["kind"], inventory, player_tile, chunks, mining_tile, inventory_number)
    elif "floor" in mining_tile and mining_tile["floor"] not in UNBREAK:
        delete_mining_tile, inventory, mining_tile = break_floor(mining_tile, inventory, inventory_number, player_tile)
    chunks[location["tile"][0], location["tile"][1]][location["tile"][2], location["tile"][3]] = player_tile
    if delete_mining_tile:
        del chunks[grid_position[0]][grid_position[1]]
    else:
        chunks[grid_position[0]][grid_position[1]] = mining_tile
    return chunks, location, machine_ui


import pygame as pg

from ...info import ATTRIBUTES, SCREEN_SIZE, UI_SCALE, SLOT_SIZE, TILE_UI_SIZE, HALF_SCREEN_SIZE
from .craft_rendering import render_craft
from .store_rendering import render_store
from .machine_rendering import render_machine

def render_open(machine_ui, window, images, recipe_number, machine_inventory):
    attributes = ATTRIBUTES.get(machine_ui, ())
    if "open" in attributes:
        window.blit(pg.transform.scale(images["big_inventory_slot"], (320 * UI_SCALE, 128 * UI_SCALE)), (HALF_SCREEN_SIZE - 160 * UI_SCALE, SCREEN_SIZE[1] - 160 * UI_SCALE))
        window.blit(pg.transform.scale(images["inventory_slot_3"], SLOT_SIZE), (HALF_SCREEN_SIZE + 88 * UI_SCALE, SCREEN_SIZE[1] - 80 * UI_SCALE))
        window.blit(pg.transform.scale(images[machine_ui], TILE_UI_SIZE), (HALF_SCREEN_SIZE + 96 * UI_SCALE, SCREEN_SIZE[1] - 76 * UI_SCALE))
        if "craft" in attributes:
            window = render_craft(window, machine_ui, images, recipe_number)
        elif "machine" in attributes:
            window = render_machine(window, machine_ui, images, machine_inventory, recipe_number)
        elif "store" in attributes:
            window = render_store(window, machine_ui, images, machine_inventory)
    return window


from ...info import RECIPES, INVENTORY_SIZE

def recipe(machine_ui: str, recipe_number: int, inventory: dict[str, int],):
    output_item, output_amount = RECIPES[machine_ui][recipe_number][0]
    input = RECIPES[machine_ui][recipe_number][1]
    if (len(inventory) >= INVENTORY_SIZE[0] and inventory.get(output_item, 0) + output_amount >= INVENTORY_SIZE[1]):
        return inventory
    for i in range(0, len(input)):
        if input[i][0] not in inventory or inventory[input[i][0]] < input[i][1]:
            return inventory
    for i in range(0, len(input)):
        inventory[input[i][0]] -= input[i][1]
        if inventory[input[i][0]] <= 0:
            del inventory[input[i][0]]
    inventory[output_item] = inventory.get(output_item, 0) + output_amount
    return inventory


import pygame as pg

from ..info import TILE_SIZE, HALF_SIZE, CHUNK_SIZE, SCREEN_SIZE, MULTI_TILES, FLOOR
from .game_rendering import render_ghost, render_mined, render_hand, render_lights, render_map

def render_tiles(
    chunks,
    location,
    zoom,
    target_zoom,
    inventory,
    inventory_number,
    tick,
    camera,
    position,
    window,
    images
):
    window.fill((206, 229, 242))
    player_pixel_position = (location["real"][2] * TILE_SIZE + location["real"][0] * CHUNK_SIZE + HALF_SIZE, location["real"][3] * TILE_SIZE + location["real"][1] * CHUNK_SIZE + HALF_SIZE)
    interpolation = max(min(abs(1 - target_zoom / zoom) * 0.5 + 0.2, 1.0), 0.0)
    camera = ((SCREEN_SIZE[0] * 5 / 8 - player_pixel_position[0] * zoom - position[0] / 4) * interpolation + camera[0] * (1 - interpolation), (SCREEN_SIZE[1] * 5 / 8 - player_pixel_position[1] * zoom - position[1] / 4) * interpolation + camera[1] * (1 - interpolation))
    scaled_image = {}
    for image in images:
        if image in FLOOR:
            scaled_image[image] = pg.transform.scale(images[image], ((TILE_SIZE + 2) * zoom, (TILE_SIZE + 2) * zoom))
        else:
            size = MULTI_TILES.get(image, (1, 1))
            scaled_image[image] = pg.transform.scale(images[image], ((TILE_SIZE * size[0] + 2) * zoom, ((size[1] + 1 / 2) * TILE_SIZE + 2) * zoom))
    window = render_map(location, chunks, camera, zoom, scaled_image, window, images)
    window = render_hand(inventory, inventory_number, camera, location, zoom, window, images)
    window = render_ghost(position, camera, zoom, chunks, location, inventory, inventory_number, scaled_image, window)
    window = render_lights(tick, chunks, location, zoom, camera, window)
    window = render_mined(location["mined"], chunks, camera, zoom, window, images)
    return camera, window


import os

import pygame as pg

from .info import FPS
from .other_systems.update import update
from .other_systems.settings_saving import settings_save
from .other_systems.rendering import render
from .other_systems.game_state import GameState


def main() -> None:
    pg.init()
    pg.mouse.set_visible(False)

    state = GameState()
    chunks = {}
    window = pg.display.set_mode((0, 0), pg.FULLSCREEN)
    SPRITES_FOLDER = os.path.normpath(os.path.join(__file__, "../..", "sprites"))
    IMAGES = {}
    for filename in os.listdir(SPRITES_FOLDER):
        IMAGES[filename.split(".")[0]] = pg.image.load(os.path.join(SPRITES_FOLDER, filename)).convert_alpha()
    while state.run:
        state.position = pg.mouse.get_pos()
        chunks = update(state, chunks)
        state.camera, window = render(state, chunks, window, IMAGES)
        state.clock.tick(FPS)
    pg.quit()
    settings_save(state.controls)


import os

import pygame as pg


SETTINGS_FILE = os.path.normpath(os.path.join(__file__, "../..", "settings.txt"))

def settings_load():
    if os.path.exists(SETTINGS_FILE):
        with open(SETTINGS_FILE, "r", encoding="utf-8") as file:
            controls = [int(i) for i in file.read().split(";")[0].split(":") if i]
    else:
        controls = [pg.K_w, pg.K_a, pg.K_s, pg.K_d, pg.K_e, pg.K_z, pg.K_x, pg.K_1, pg.K_2, pg.K_3, pg.K_4, pg.K_5, pg.K_6, pg.K_7, pg.K_8, pg.K_9, pg.K_0, pg.K_PLUS, pg.K_LSHIFT, pg.K_RIGHT, pg.K_LEFT, pg.K_ESCAPE, pg.K_DOWN, pg.K_UP, pg.K_LCTRL]
    return controls

def settings_save(controls):
    control_str = ""
    for i in controls:
        control_str += f"{i}:"
    with open(SETTINGS_FILE, "w", encoding="utf-8") as file:
        file.write(f"{control_str}")


from ...info import SCREEN_SIZE, UI_SCALE, INVENTORY_SIZE, RECIPES, ATTRIBUTES, MACHINES, VALUES
from .put_in import put_in
from .take_out import take_out

def machine_storage(position, chunks, location, inventory, machine_ui):
    if "inventory" not in chunks[location["opened"][0]][location["opened"][1]]:
        chunks[location["opened"][0]][location["opened"][1]]["inventory"] = {}
    moved_x = position[0] - SCREEN_SIZE[0] // 2
    machine = chunks[location["opened"][0]][location["opened"][1]]
    machine_recipe = RECIPES[machine_ui][machine.get("recipe", 0)]
    holding_over_inventory = position[1] >= SCREEN_SIZE[1] - 32 * UI_SCALE and abs(moved_x) <= 16 * INVENTORY_SIZE[0] * UI_SCALE
    if holding_over_inventory:
        inventory_number = ((moved_x - 16 * UI_SCALE * (INVENTORY_SIZE[0] % 2)) // (32 * UI_SCALE) + INVENTORY_SIZE[0] // 2 + INVENTORY_SIZE[0] % 2)
        if inventory_number < len(inventory):
            item = list(inventory.items())[((moved_x - 16 * UI_SCALE * (INVENTORY_SIZE[0] % 2)) // (32 * UI_SCALE) + INVENTORY_SIZE[0] // 2 + INVENTORY_SIZE[0] % 2)]
            may_put_in = False
            convert = False
            for i in range(0, len(machine_recipe[1])):
                if machine_recipe[1][i][0] == item[0]:
                    may_put_in = True
            if "value" in ATTRIBUTES[machine["kind"]]:
                if item[0] in VALUES[MACHINES[machine["kind"]]]:
                    may_put_in = True
                    convert = True
            if may_put_in:
                chunks = put_in(chunks, location, inventory, machine_ui, moved_x, machine["inventory"])
                if convert:
                    chunks[location["opened"][0]][location["opened"][1]]["inventory"][MACHINES[machine["kind"]]] = VALUES[MACHINES[machine["kind"]]][item[0]] * machine["inventory"][item[0]]
                    del chunks[location["opened"][0]][location["opened"][1]]["inventory"][item[0]]
    slot_row = (position[1] - SCREEN_SIZE[1] + 144 * UI_SCALE) // (32 * UI_SCALE)
    if slot_row == 2 and (moved_x + 112 * UI_SCALE) // (32 * UI_SCALE) == 0:
        item = (machine_recipe[0][0], machine["inventory"].get(machine_recipe[0][0], 0))
        if machine["inventory"].get(item[0], 0) > 0:
            chunks = take_out(chunks, location, inventory, item)
    return chunks


import os

import pygame as pg

from ..info import SCREEN_SIZE, WORLD_TYPES

pg.font.init()

MENU_FONT = pg.font.SysFont("Lucida Console", 50)
CONTROL_NAMES = ["Move up ", "Move left ", "Move down ", "Move right", "Inventory ", "Zoom in", "Zoom out", "Slot 1", "Slot 2", "Slot 3", "Slot 4", "Slot 5", "Slot 6", "Slot 7", "Slot 8", "Slot 9", "Slot 10", "Slot 11", "Slot 12", "Hotbar scroll right", "Hotbar scroll left", "Go to menu", "Menu scroll down", "Menu scroll up", "Sneak"]
SAVES_FOLDER = os.path.normpath(os.path.join(__file__, "../../..", "saves"))

if not os.path.exists(SAVES_FOLDER):
    os.makedirs(SAVES_FOLDER)

def render_menu(
    menu_placement: str,
    save_file_name: str,
    controls: list,
    window,
    scroll,
    control_adjusted,
    world_type,
):
    window.fill((206, 229, 242))
    if menu_placement == "load_save":
        window.blit(MENU_FONT.render("Back to menu", False, (19, 17, 18)), (0, 0))
        window.blit(MENU_FONT.render("Create new world", False, (19, 17, 18)), (0, 50))
        saves = [f[:-len(".txt")] for f in os.listdir(SAVES_FOLDER)]
        for i in range(0, len(saves)):
            window.blit(MENU_FONT.render(f"[x] [{saves[i]}]", False, (19, 17, 18)), (0, 100 + i * 50))
    elif menu_placement == "save_creation":
        window.blit(MENU_FONT.render("Name your new save?", False, (19, 17, 18)), (0, 0))
        window.blit(MENU_FONT.render(save_file_name, False, (19, 17, 18)), (0, 100))
        window.blit(MENU_FONT.render("Proceed", False, (19, 17, 18)), (0, 200))
        window.blit(MENU_FONT.render("Don't save", False, (19, 17, 18)), (0, 300))
    elif menu_placement == "save_options":
        window.blit(MENU_FONT.render("Create new save", False, (19, 17, 18)), (0, 0))
        window.blit(MENU_FONT.render(f"World type: {WORLD_TYPES[world_type].capitalize()}", False, (19, 17, 18)), (0, 50))
    elif menu_placement.split("_")[0] == "options":
        if menu_placement == "options_game":
            window.blit(MENU_FONT.render("Return to game", False, (19, 17, 18)), (0, 0))
            window.blit(MENU_FONT.render("Save and Quit", False, (19, 17, 18)), (0, 100))
        elif menu_placement == "options_main":
            window.blit(MENU_FONT.render("Back to menu", False, (19, 17, 18)), (0, 0))
        window.blit(MENU_FONT.render("Controls options", False, (19, 17, 18)), (0, 200))
    elif menu_placement == "main_menu":
        window.blit(MENU_FONT.render("Play", False, (19, 17, 18)), (0, 0))
        window.blit(MENU_FONT.render("Options", False, (19, 17, 18)), (0, 100))
        window.blit(MENU_FONT.render("Quit Game", False, (19, 17, 18)), (0, 200))
    elif menu_placement == "controls_options":
        pg.draw.rect(window, (181, 102, 60), pg.Rect(0, 50 * (control_adjusted - scroll), SCREEN_SIZE[0], 50))
        for y in range(0, len(controls)):
            window.blit(MENU_FONT.render(f"{CONTROL_NAMES[y]}: {pg.key.name(controls[y]).capitalize()}", False, (19, 17, 18)), (0, 50 * (y - scroll)))
    return window


from .point import left, up
from .rabbit import rabbit_hole, rabbit_entity
from .machine import machine
from .growth import grow
from ...info import ATTRIBUTES

def update_tile(current_tile, chunks, chunk, tile, delete_tiles, create_tiles, tick):
    if "grow" in ATTRIBUTES.get(current_tile["kind"], ()):
        chunks[chunk][tile] = grow(current_tile)
    elif current_tile["kind"] == "left":
        chunks, delete_tiles = left(chunks, chunk, tile, delete_tiles)
    elif current_tile["kind"] == "up":
        chunks, delete_tiles = up(chunks, chunk, tile, delete_tiles)
    elif current_tile["kind"] == "rabbit hole":
        chunks, create_tiles = rabbit_hole(chunks, chunk, tile, current_tile, create_tiles)
    elif "rabbit" in ATTRIBUTES.get(current_tile["kind"], ()):
        create_tiles, delete_tiles = rabbit_entity(chunks, chunk, tile, current_tile, create_tiles, delete_tiles)
    elif "machine" in ATTRIBUTES.get(current_tile["kind"], ()):
        chunks = machine(chunks, chunk, tile, current_tile, tick)
    return chunks, create_tiles, delete_tiles


from ...info import TILE_SIZE, ATTRIBUTES, RECIPES, INVENTORY_SIZE
from .right_click_updates import right_click
from .left_click_update import left_click

def button_press(button, position, zoom, chunks, location, machine_ui, inventory, health, machine_inventory, tick, inventory_number, recipe_number, camera):
    world_x = int((position[0] - camera[0]) // (TILE_SIZE * zoom))
    world_y = int((position[1] - camera[1]) // (TILE_SIZE * zoom))
    if (world_x - location["tile"][0] * 16 - location["tile"][2]) ** 2 + (world_y - location["tile"][1] * 16 - location["tile"][3]) ** 2 <= 20 or "open" in ATTRIBUTES.get(machine_ui, ()):
        grid_position = [(world_x // 16, world_y // 16), (world_x % 16, world_y % 16)]
        if grid_position[1] in chunks[grid_position[0]] and "kind" in chunks[grid_position[0]][grid_position[1]]:
            while "point" in ATTRIBUTES.get(chunks[grid_position[0]][grid_position[1]]["kind"], ()):
                if chunks[grid_position[0]][grid_position[1]]["kind"] == "left":
                    grid_position = [(grid_position[0][0] - (grid_position[1][0] == 0), grid_position[0][1]), ((grid_position[1][0] - 1) % 16, grid_position[1][1])]
                elif chunks[grid_position[0]][grid_position[1]]["kind"] == "up":
                    grid_position = [(grid_position[0][0], grid_position[0][1] - (grid_position[1][1] == 0)), (grid_position[1][0], (grid_position[1][1] - 1) % 16)]
                    
        if button == 1:
            machine_ui, chunks, location, machine_inventory, tick = left_click(machine_ui, grid_position, chunks, inventory_number, health, position, recipe_number, location, inventory, machine_inventory, tick)
        elif button == 3:
            chunks, location, machine_ui = right_click(chunks, grid_position, inventory, inventory_number, location, machine_ui)
    
    if button == 4 or button == 5:
        if "craft" in ATTRIBUTES.get(machine_ui, ()) or "machine" in ATTRIBUTES.get(machine_ui, ()):
            chunks[location["opened"][0]][location["opened"][1]]["recipe"] = (recipe_number + (button == 5) - (button == 4)) % len(RECIPES[machine_ui])
        else:
            inventory_number = (inventory_number + (button == 5) - (button == 4)) % INVENTORY_SIZE[0]
    return chunks, location, machine_ui, machine_inventory, tick, inventory_number


from random import choice

def find_empty_place(tile, chunk, chunks):
    empty_places = []
    for x in range(-1, 2):
        for y in range(-1, 2):
            tile_pos = ((tile[0] + x) % 16, (tile[1] + y) % 16)
            chunk_pos = (chunk[0] + (tile[0] + x) // 16, chunk[1] + (tile[1] + y) // 16)
            if tile_pos not in chunks.get(chunk_pos, {}) or "kind" not in chunks[chunk_pos][tile_pos]:
                empty_places.append((x, y))
    return choice(empty_places) if empty_places else None


from random import choice, randint

from .find_empty_place import find_empty_place

def rabbit_hole(chunks, chunk, tile, current_tile, create_tiles):
    if randint(0, 10000) == 0:
        spawn_pos = (chunk[0] * 16 + tile[0], chunk[1] * 16 + tile[1])
        animal = choice(({"kind": "rabbit adult", "inventory": {"rabbit meat": 2, "rabbit fur": 1}, "spawn": spawn_pos}, {"kind": "rabbit child", "spawn": spawn_pos}))
        if animal["kind"] in current_tile["inventory"]:
            empty = find_empty_place(tile, chunk, chunks)
            if empty:
                x, y = empty
                current_tile["inventory"][animal["kind"]] -= 1
                if current_tile["inventory"][animal["kind"]] <= 0:
                    del current_tile["inventory"][animal["kind"]]
                create_tiles.append((
                    (chunk[0] + (tile[0] + x) // 16, chunk[1] + (tile[1] + y) // 16),
                    ((tile[0] + x) % 16, (tile[1] + y) % 16),
                    animal
                ))
    return chunks, create_tiles

def rabbit_entity(chunks, chunk, tile, current_tile, create_tiles, delete_tiles):
    if randint(0, 100) == 0:
        empty = find_empty_place(tile, chunk, chunks)
        if empty:
            x, y = empty
            if abs(chunk[0] * 16 + tile[0] + x - current_tile["spawn"][0]) <= 8 and \
                abs(chunk[1] * 16 + tile[1] + y - current_tile["spawn"][1]) <= 8:
                create_tiles.append((
                    (chunk[0] + (tile[0] + x) // 16, chunk[1] + (tile[1] + y) // 16),
                    ((tile[0] + x) % 16, (tile[1] + y) % 16),
                    {"kind": current_tile["kind"], "inventory": current_tile["inventory"], "spawn": current_tile["spawn"]}
                ))
                delete_tiles.append((chunk, tile))
    return create_tiles, delete_tiles


from ...info import ATTRIBUTES

def unlock(inventory, inventory_number, chunks, grid_position):
    if len(inventory) > inventory_number:
        inventory_key = list(inventory.keys())[inventory_number]
        if "key" in ATTRIBUTES.get(inventory_key, ()):
            if inventory_key.split()[0] == chunks[grid_position[0]][grid_position[1]]["kind"].split()[0]:
                inventory[inventory_key] -= 1
                if inventory[inventory_key] == 0:
                    del inventory[inventory_key]
                del chunks[grid_position[0]][grid_position[1]]
    return chunks


from noise import pnoise2

from ..info import BIOMES


def determine_biome(world_x, world_y, noise_offset):
    biome_value = pnoise2(world_x / 120 + noise_offset[0], world_y / 120 + noise_offset[1], 3, 0.5, 2)
    biome = "plains"
    for noise_chunk in BIOMES:
        if noise_chunk[0] < biome_value < noise_chunk[1]:
            biome = noise_chunk[2]
            break
    return biome


from json import dumps
from os import path

from ...render.menu_rendering import SAVES_FOLDER
from .load_save import save_loading
from .create_save import save_creating
from .options import option
from ...info import SCREEN_SIZE, WORLD_TYPES

def update_mouse(state, event, chunks):
    if state.menu_placement == "load_save":
        if state.position[1] <= 50:
            state.menu_placement = "main_menu"
        elif state.position[1] <= 100:
            state.menu_placement = "save_options"
            state.world_type = 0
        else:
            chunks = save_loading(state, chunks)
    elif state.menu_placement.startswith("options"):
        option(state, chunks)

    elif state.menu_placement == "save_options":
        if state.position[1] <= 50:
            chunks = save_creating(state, chunks)
        elif state.position[1] <= 100:
            state.world_type = (state.world_type + 1) % len(WORLD_TYPES)
    elif state.menu_placement == "save_creation":
        if 200 <= state.position[1] <= 250 and state.save_file_name != "" and state.save_file_name.split("_")[0] != "autosave":
            state.menu_placement = "main_menu"
            with open(path.join(SAVES_FOLDER, state.save_file_name + ".txt"), "w", encoding="utf-8") as file:
                file.write(f"{chunks};{state.location['tile']};{state.tick};{state.noise_offset};{state.world_type}")
            state.save_file_name = ""
            chunks = {}
        elif 300 <= state.position[1] <= 350:
            state.menu_placement = "main_menu"
            state.save_file_name = ""
            chunks = {}

    elif state.menu_placement == "main_menu":
        if 0 <= state.position[1] <= 50:
            state.menu_placement = "load_save"
        elif 100 <= state.position[1] <= 150:
            state.menu_placement = "options_main"
        elif 200 <= state.position[1] <= 250:
            state.run = False

    elif state.menu_placement == "controls_options":
        if event.button == 4:
            if state.scroll > 0:
                state.scroll -= 1
        elif event.button == 5:
            if state.scroll < len(state.controls) - SCREEN_SIZE[1] // 50 - 1:
                state.scroll += 1
        else:
            state.control_adjusted = state.scroll + state.position[1] // 50
    return chunks


from random import random

from ...info import GROW_CHANCE, GROW_TILES, FPS

def grow(tile, guarantee = False):
    if random() < 1 / (GROW_CHANCE[tile["kind"]] * FPS) or guarantee:
        floor = tile["floor"]
        if "spawn" in tile:
            spawn = tile["spawn"]
            tile = GROW_TILES[tile["kind"]]
            tile["spawn"] = spawn
        else:
            tile = GROW_TILES[tile["kind"]]
        tile["floor"] = floor
    return tile


def put_in(chunks, location, inventory, machine_storage, slot_number, machine_inventory):
    if slot_number < len(inventory):
        item = list(inventory.items())[slot_number]
        machine_item = machine_inventory.get(item[0], 0)
        if not (machine_item == 0 and len(machine_inventory) == machine_storage[0]):
            if machine_item + item[1] <= machine_storage[1]:
                chunks[location["opened"][0]][location["opened"][1]]["inventory"][item[0]] = machine_item + item[1]
                del chunks[location["tile"][0], location["tile"][1]][location["tile"][2], location["tile"][3]]["inventory"][item[0]]
            else:
                chunks[location["opened"][0]][location["opened"][1]]["inventory"][item[0]] = machine_storage[1]
                chunks[location["tile"][0], location["tile"][1]][location["tile"][2], location["tile"][3]]["inventory"][item[0]] = (machine_item + item[1] - machine_storage[1])
    return chunks


import pygame as pg

from ..render import render_menu, render_tiles, render_ui
from .game_state import GameState


def render(state: GameState, chunks, window, images) -> tuple:
    if state.menu_placement != "main_game":
        window = render_menu(state.menu_placement, state.save_file_name, state.controls, window, state.scroll, state.control_adjusted, state.world_type)
        window.blit(pg.transform.scale(images["cursor"], (32, 32)), (state.position[0] - 16, state.position[1] - 16))
    else:
        state.camera, window = render_tiles(
            chunks,
            state.location,
            state.zoom,
            state.target_zoom,
            state.inventory,
            state.inventory_number,
            state.tick,
            state.camera,
            state.position,
            window,
            images
        )
        window = render_ui(
            state.inventory_number,
            state.inventory,
            state.machine_ui,
            chunks[state.location["opened"][0]].get(state.location["opened"][1], chunks[0, 0][0, 0]).get("recipe", 0),
            state.health,
            state.machine_inventory,
            window,
            images,
        )
        cursor_size = (int(32 * state.zoom), int(32 * state.zoom))
        cursor_place = (state.position[0] - 16 * state.zoom, state.position[1] - 16 * state.zoom)
        window.blit(pg.transform.scale(images["cursor"], cursor_size), cursor_place)
    pg.display.update()
    return state.camera, window


from ...info import TILE_SIZE, HALF_SIZE, SCREEN_SIZE, MULTI_TILES, FLOOR

def render_ghost(position, camera, zoom, chunks, location, inventory, inventory_number, scaled_image, window):
    world_x = int((position[0] - camera[0]) // (TILE_SIZE * zoom))
    world_y = int((position[1] - camera[1]) // (TILE_SIZE * zoom))
    grid_position = [(world_x // 16, world_y // 16), (world_x % 16, world_y % 16)]
    if inventory_number < len(inventory):
        if grid_position[0] in chunks:
            if (world_x - location["tile"][0] * 16 - location["tile"][2]) ** 2 + (world_y - location["tile"][1] * 16 - location["tile"][3]) ** 2 <= 20:
                inventory_key = list(inventory.keys())[inventory_number]
                size = MULTI_TILES.get(inventory_key, (1, 1))
                if -TILE_SIZE * zoom * size[0] <= world_x <= SCREEN_SIZE[0] and -TILE_SIZE * zoom * size[1] * 3 / 2 <= world_y <= SCREEN_SIZE[1] and inventory_number < len(inventory):
                    placement = (camera[0] + world_x * TILE_SIZE * zoom, camera[1] + (world_y * TILE_SIZE - HALF_SIZE) * zoom,)
                    alpha_image = scaled_image[inventory_key].copy()
                    alpha_image.set_alpha(85)
                    if inventory_key in FLOOR:
                        window.blit(alpha_image, (placement[0], placement[1] + HALF_SIZE * zoom))
                    else:
                        window.blit(alpha_image, placement)
    return window


from ...info import INVENTORY_SIZE, HEALTH, ATTRIBUTES
from .damage_calculation import calculate_damage

def break_tile(mining_kind, inventory, player_tile, chunks, mining_tile, inventory_number):
    delete_mining_tile = False
    if mining_kind != "player":
        if "health" not in mining_tile:
            mining_tile["health"] = HEALTH[mining_tile["kind"]]
        mining_tile["health"] -= calculate_damage(mining_tile["kind"], inventory, inventory_number)
        if mining_tile["health"] <= 0:
            mining_floor_exist = "floor" in mining_tile
            if mining_floor_exist:
                mining_floor = mining_tile["floor"]
            if mining_kind != "player":
                junk_inventory = {}
                if "no_pickup" not in ATTRIBUTES.get(mining_tile["kind"], ()):
                    inventory[mining_kind] = inventory.get(mining_kind, 0) + 1
                if "inventory" in mining_tile:
                    for item, amount in mining_tile["inventory"].items():
                        if item in player_tile["inventory"]:
                            player_tile["inventory"][item] += amount
                            if inventory[item] > INVENTORY_SIZE[1]:
                                junk_inventory[item] = player_tile["inventory"][item] - INVENTORY_SIZE[1]
                                player_tile["inventory"][item] = INVENTORY_SIZE[1]
                        else:
                            if len(inventory) < INVENTORY_SIZE[0]:
                                player_tile["inventory"][item] = amount
                            else:
                                junk_inventory[item] = amount
                if mining_floor_exist:
                    mining_tile = {}
                else:
                    delete_mining_tile = True
                if len(junk_inventory) > 0:
                    mining_tile = {"kind": "junk", "inventory": junk_inventory}
            if mining_floor_exist:
                mining_tile["floor"] = mining_floor
    return chunks, delete_mining_tile, mining_tile


NOISE_STRUCTURES = {
    "forest": (
        ((0, 0.05), "mushroom hut"),
    ),
    "plains": (
        ((0, 0), "copper dungeon"),
    ),
}
ROOM_COLORS = {
    (247, 247, 255): {"kind": "mushroom block", "floor": "mushroom floor"},
    (138, 138, 140): {"floor": "mushroom floor"},
    (53, 53, 54): {"floor": "mushroom door"},
    (106, 228, 138): {"kind": "mushroom shaper", "floor": "mushroom floor"},
    (92, 74, 49): {"kind": "small crate", "loot": "mushroom chest", "floor": "mushroom floor"},
    (136, 68, 31): {"kind": "copper brick", "floor": "copper brick floor"},
    (181, 102, 60): {"floor": "copper brick floor"},
}
STRUCTURE_ENTRANCE = {"copper dungeon": {"kind": "glass lock"}, "mushroom hut": {"floor": "mushroom door"}}
STRUCTURE_ROOM_SIZES = {"copper dungeon": ((1, 1), (2, 1), (2, 2)), "mushroom hut": ((1, 1),)}
STRUCTURE_SIZE = {"copper dungeon": 0.8, "mushroom hut": 0}
STRUCTURE_ROOMS = {
    "copper dungeon": {
        (1, 1): ("treasury", "hallway"),
        (2, 1): ("library", "banquet"),
        (2, 2): ("forge",),
    },
    "mushroom hut": {
        (1, 1): ("mushroom hut",),
    },
}
LOOT_TABLES = {
    "mushroom chest": {
        (0.7, "spore", 1, 5),
        (0.6, "mushroom", 2, 7),
        (0.4, "mushroom block", 3, 5),
        (0.25, "fertilizer", 1, 2),
        (0.2, "mushroom floor", 2, 3),
        (0.15, "plant bouquet", 1, 3),
        (0.1, "mushroom shaper", 1, 1),
        (0.05, "mushroom door", 1, 2),
        (0.03, "composter", 1, 1),
        (0.02, "bonsai pot", 1, 2),
    }
}


from ...info import TILE_SIZE, HALF_SIZE, CHUNK_SIZE, SCREEN_SIZE, MULTI_TILES, ATTRIBUTES
from .border_rendering import render_border

def render_map(location, chunks, camera, zoom, scaled_image, window, images):
    for chunk_y in range(-3, 4):
        for chunk_x in range(-3, 4):
            chunk = (chunk_x + location["tile"][0], chunk_y + location["tile"][1])
            if chunk in chunks:
                for y in range(0, 16):
                    for x in range(0, 16):
                        tile = (x, y)
                        if tile in chunks[chunk]:
                            current_tile = chunks[chunk][tile]
                            placement = (camera[0] + (x * TILE_SIZE + chunk[0] * CHUNK_SIZE) * zoom, camera[1] + (y * TILE_SIZE + chunk[1] * CHUNK_SIZE) * zoom,)
                            if -TILE_SIZE * zoom <= placement[0] <= SCREEN_SIZE[0] and -TILE_SIZE * zoom <= placement[1] <= SCREEN_SIZE[1]:
                                if "floor" in chunks[chunk][tile]:
                                    window.blit(scaled_image[current_tile["floor"]], placement)
                                    render_border(chunk, x, y, chunks, placement, zoom, window, current_tile)
    for chunk_y in range(-3, 4):
        for chunk_x in range(-3, 4):
            chunk = (chunk_x + location["tile"][0], chunk_y + location["tile"][1])
            if chunk in chunks:
                for y in range(0, 16):
                    for x in range(0, 16):
                        tile = (x, y)
                        if tile in chunks[chunk] and "kind" in chunks[chunk][tile] and "point" not in ATTRIBUTES.get(chunks[chunk][tile]["kind"], ()):
                            current_tile = chunks[chunk][tile]
                            placement = (camera[0] + (x * TILE_SIZE + chunk[0] * CHUNK_SIZE) * zoom, camera[1] + (y * TILE_SIZE + chunk[1] * CHUNK_SIZE - HALF_SIZE) * zoom,)
                            size = MULTI_TILES.get(current_tile["kind"], (1, 1))
                            if -TILE_SIZE * zoom * size[0] <= placement[0] <= SCREEN_SIZE[0] and -TILE_SIZE * zoom * size[1] * 3 / 2 <= placement[1] <= SCREEN_SIZE[1]:
                                if isinstance(chunks[chunk][tile]["kind"], str):
                                    window.blit(scaled_image[current_tile["kind"]], placement)
                                if "table" in ATTRIBUTES.get(current_tile["kind"], ()) and "inventory" in current_tile:
                                    window.blit(scaled_image[list(current_tile["inventory"])[0]], (placement[0], placement[1] - HALF_SIZE * zoom * (images[list(current_tile["inventory"])[0]].get_size()[1] // 8 - 2)))
    return window


ATTRIBUTES = {
    "big rock": {"multi", "no_pickup"},
    "bonsai pot": {"open", "machine"},
    "campfire": {"open", "craft", "light"},
    "carrot": {"grow", "eat"},
    "carroot": {"no_pickup"},
    "coal ore": {"no_pickup"},
    "compost": {"fertilize"},
    "composter": {"open", "machine", "value"},
    "fertilizer": {"fertilize"},
    "glass key": {"key"},
    "glass lock": {"lock"},
    "manual press": {"open", "craft", "multi"},
    "masonry bench": {"open", "craft", "multi"},
    "mushroom": {"eat"},
    "mushroom shaper": {"open", "craft"},
    "mushroom stew": {"eat"},
    "rabbit adult": {"rabbit", "no_pickup"},
    "rabbit child": {"grow", "rabbit", "no_pickup"},
    "rabbit meat": {"eat"},
    "roasted mushroom": {"eat"},
    "roasted rabbit meat": {"eat"},
    "sapling": {"grow"},
    "sawbench": {"open", "craft", "multi"},
    "small barrel": {"open", "store"},
    "small crate": {"open", "store"},
    "spore": {"grow"},
    "tree": {"no_pickup"},
    "treeling": {"grow", "no_pickup"},
    "wooden bed": {"sleep", "multi"},
    "wooden table": {"store", "table"},
    "workbench": {"open", "craft"},

    # Special tiles
    "left": {"point", "no_pickup"},
    "up": {"point", "no_pickup"},
    "player": {"open", "craft", "no_pickup"},
    "junk": {"no_pickup"},
    "corpse": {"no_pickup"},
    "obelisk": {"multi", "no_pickup"},
}


import pygame as pg

from ...info import SCREEN_SIZE, UI_SCALE, UI_FONT, HEALTH

def render_health(window, images, health):
    max_health = HEALTH["player"]
    window.blit(pg.transform.scale(images["health_bar"], (128 * UI_SCALE, 32 * UI_SCALE)), (SCREEN_SIZE[0] - 128 * UI_SCALE, 0))
    window.blit(pg.transform.scale(images["health_end"], (16 * UI_SCALE, 16 * UI_SCALE)), (SCREEN_SIZE[0] + (health * 64 / max_health - 96) * UI_SCALE, 8 * UI_SCALE))
    pg.draw.rect(window, (181, 102, 60), pg.Rect(SCREEN_SIZE[0] - 96 * UI_SCALE, 8 * UI_SCALE, health * 64 * UI_SCALE / max_health, 16 * UI_SCALE))
    window.blit(UI_FONT.render(f"{health} / {max_health}", False, (206, 229, 242)), (SCREEN_SIZE[0] - 80 * UI_SCALE, 12 * UI_SCALE))
    return window


import pygame as pg

from .settings_saving import settings_load


class GameState:
    def __init__(self):
        self.controls = settings_load()
        self.clock = pg.time.Clock()
        self.location = {"mined": ((0, 0), (0, 2)), "opened": ((0, 0), (0, 0))}
        self.run = True
        self.zoom = 1
        self.target_zoom = 1
        self.inventory_number = 0
        self.inventory = {}
        self.health = 20
        self.save_file_name = ""
        self.menu_placement = "main_menu"
        self.velocity = [0, 0]
        self.machine_ui = "game"
        self.control_adjusted = -1
        self.scroll = 0
        self.machine_inventory = {}
        self.camera = (0, 0)
        self.checked = set()
        self.world_type = 0


from random import random, randint

from ..info import LOOT_TABLES


def calculate_loot(tile):
    loot_table = LOOT_TABLES[tile["loot"]]
    tile["inventory"] = {}
    for item in loot_table:
        if random() < item[0]:
            tile["inventory"][item[1]] = randint(item[2], item[3])
    if len(tile["inventory"]) <= 1:
        tile = calculate_loot(tile)
    return tile


from ...other_systems.game_state import GameState
from ..chunk_update import update_tile, create_tile, delete_tile

def update_tiles(state: GameState, chunks):
    delete_tiles = []
    create_tiles = []
    tile_location = state.location["tile"]

    for chunk_dx in range(-3, 4):
        for chunk_dy in range(-3, 4):
            chunk = (chunk_dx + tile_location[0], chunk_dy + tile_location[1])
            if chunk in chunks:
                for tile in list(chunks[chunk]):
                    current_tile = chunks[chunk][tile]
                    if "kind" in current_tile:
                        chunks, create_tiles, delete_tiles = update_tile(current_tile, chunks, chunk, tile, delete_tiles, create_tiles, state.tick)

    chunks = create_tile(chunks, create_tiles)
    chunks = delete_tile(chunks, delete_tiles)
    return chunks


def create_tile(chunks, create_tiles):
    for chunk_pos, tile_pos, tile_data in create_tiles:
        chunk_tiles = chunks.setdefault(chunk_pos, {})
        if tile_pos in chunk_tiles:
            current_tile = chunk_tiles[tile_pos]
            tile_data["floor"] = current_tile["floor"]
        chunk_tiles[tile_pos] = tile_data
    return chunks

def delete_tile(chunks, delete_tiles):
    for chunk_pos, tile_pos in delete_tiles:
        tile = chunks[chunk_pos][tile_pos]
        if "floor" in tile:
            chunks[chunk_pos][tile_pos] = {"floor": tile["floor"]}
        else:
            del chunks[chunk_pos][tile_pos]
    return chunks


import pygame as pg

from ...info import SCREEN_SIZE, UI_SCALE, UI_FONT, INVENTORY_SIZE, SLOT_SIZE, TILE_UI_SIZE, FLOOR_SIZE, HALF_SCREEN_SIZE, FLOOR

def render_inventory(inventory_number, window, images, inventory):
    for i in range(0, INVENTORY_SIZE[0]):
        if i == inventory_number:
            window.blit(pg.transform.scale(images["inventory_slot_2"], SLOT_SIZE), (HALF_SCREEN_SIZE + (32 * i - 16 * INVENTORY_SIZE[0]) * UI_SCALE, SCREEN_SIZE[1] - 32 * UI_SCALE))
        else:
            window.blit(pg.transform.scale(images["inventory_slot"], SLOT_SIZE), (HALF_SCREEN_SIZE + (32 * i - 16 * INVENTORY_SIZE[0]) * UI_SCALE, SCREEN_SIZE[1] - 32 * UI_SCALE))
    t = 0
    for item in inventory:
        if item not in FLOOR:
            window.blit(pg.transform.scale(images[item], TILE_UI_SIZE), (HALF_SCREEN_SIZE + (32 * t - 16 * INVENTORY_SIZE[0] + 8) * UI_SCALE, SCREEN_SIZE[1] - 28 * UI_SCALE))
        else:
            window.blit(pg.transform.scale(images[item], FLOOR_SIZE), (HALF_SCREEN_SIZE + (32 * t - 16 * INVENTORY_SIZE[0] + 8) * UI_SCALE, SCREEN_SIZE[1] - 20 * UI_SCALE))
        window.blit(UI_FONT.render(str(inventory[item]), False, (19, 17, 18)), (HALF_SCREEN_SIZE + (32 * t - 16 * INVENTORY_SIZE[0] + 4) * UI_SCALE, SCREEN_SIZE[1] - 24 * UI_SCALE))
        t += 1
    return window


from .ghost_rendering import render_ghost
from .hand_rendering import render_hand
from .light_rendering import render_lights
from .map_rendering import render_map
from .mine_rendering import render_mined


from ...info.tile_info import PROCESSING_TIME
from ..left_click import recipe

def machine(chunks, chunk, tile, current_tile, tick):
    if tick % PROCESSING_TIME[current_tile["kind"]] == 0:
        chunks[chunk][tile]["inventory"] = recipe(current_tile["kind"], current_tile.get("recipe", 0), current_tile.get("inventory", {}))
    return chunks


import pygame as pg

from .user_interface_rendering import render_health, render_inventory, render_open

def render_ui(inventory_number, inventory, machine_ui, recipe_number, health, machine_inventory, window, images):
    window = render_health(window, images, health)
    window = render_inventory(inventory_number, window, images, inventory)
    window = render_open(machine_ui, window, images, recipe_number, machine_inventory)
    return window


from ..info import MULTI_TILES, ATTRIBUTES, FLOOR_TYPE

def is_placable(kind, grid_position, chunks):
    tile_size = MULTI_TILES.get(kind, (1, 1))
    for x in range(0, tile_size[0]):
        for y in range(0, tile_size[1]):
            tile_coord = (int((grid_position[1][0] + x) % 16), int((grid_position[1][1] + y) % 16))
            chunk_coord = (grid_position[0][0] + (grid_position[1][0] + x) // 16, grid_position[0][1] + (grid_position[1][1] + y) // 16)
            if tile_coord in chunks[chunk_coord]:
                current_tile = chunks[chunk_coord][tile_coord]
                tile_floor_type = FLOOR_TYPE.get(current_tile["floor"])
                if tile_floor_type == "block" or tile_floor_type == "fluid" or "grow" in ATTRIBUTES.get(kind, ()) and tile_floor_type != "soil" or "kind" in current_tile:
                    return False
            elif "grow" in ATTRIBUTES.get(kind, ()):
                return False
    return True


RESISTANCE = {
    "big rock": 2,
    "bonsai pot": 2,
    "brick": 1,
    "clay": 1,
    "coal": 1,
    "coal ore": 2,
    "composter": 1,
    "copper brick": 3,
    "copper brick floor": 3,
    "ice": 1,
    "manual press": 1,
    "masonry bench": 1,
    "mushroom block": 1,
    "mushroom door": 1,
    "mushroom floor": 1,
    "mushroom shaper": 2,
    "pebble": 1,
    "rabbit adult": 1,
    "rock": 1,
    "sawbench": 1,
    "stone": 1,
    "stone brick": 1,
    "tree": 1,
}


from random import random

from ...info import FERTILIZER_SPAWN, FERTILIZER_EFFICIENCY, GROW_CHANCE
from ...other_systems.tile_placement import place_tile
from ..chunk_update.growth import grow


def fertilize_spawn(chunks, inventory, inventory_number, grid_position):
    inventory_key = list(inventory.keys())[inventory_number]
    inventory[inventory_key] -= 1
    if inventory[inventory_key] == 0:
        del inventory[inventory_key]
    spawn_number = random()
    spawn = None
    for i in FERTILIZER_SPAWN:
        if spawn_number < i[0]:
            spawn = i[1]
            break
    if isinstance(spawn, str):
        chunks = place_tile(spawn, grid_position, chunks)
    return chunks

def fertilize_grow(chunks, inventory, inventory_number, grid_position):
    inventory_key = list(inventory.keys())[inventory_number]
    current_tile = chunks[grid_position[0]][grid_position[1]]
    if random() < 1 / GROW_CHANCE[current_tile["kind"]] * FERTILIZER_EFFICIENCY[inventory_key]:
        chunks[grid_position[0]][grid_position[1]] = grow(current_tile, True)
    inventory[inventory_key] -= 1
    if inventory[inventory_key] == 0:
        del inventory[inventory_key]
    return chunks


import pygame as pg

from ...info import TILE_SIZE, CHUNK_SIZE, HEALTH

def render_mined(location, chunks, camera, zoom, window, images):
    if location[0] in chunks and location[1] in chunks[location[0]]:
        placement = (camera[0] + (location[1][0] * TILE_SIZE + location[0][0] * CHUNK_SIZE) * zoom, camera[1] + (location[1][1] * TILE_SIZE + location[0][1] * CHUNK_SIZE + 60) * zoom,)
        current_tile = chunks[location[0]][location[1]]
        if "health" in current_tile:
            if "kind" in chunks[location[0]][location[1]]:
                max_health = HEALTH[current_tile["kind"]]
            else:
                max_health = HEALTH[current_tile["floor"]]
            window.blit(pg.transform.scale(images["tiny_bar"], (TILE_SIZE * zoom, 16 * zoom)), placement)
            if current_tile["health"] // max_health == 0:
                break_image = images[f"break_{8 - (current_tile["health"] * 8 // max_health)}"]
                break_image.set_alpha(127)
                window.blit(pg.transform.scale(break_image, (TILE_SIZE * zoom, TILE_SIZE * zoom)), (placement[0], placement[1] - TILE_SIZE * zoom))
            pg.draw.rect(window, (181, 102, 60), pg.Rect(placement[0] + 4 * zoom, placement[1] + 4 * zoom, current_tile["health"] * 44 * zoom / max_health, 8 * zoom))
    return window


from random import random

from ..info import NOISE_STRUCTURES, STRUCTURE_ENTRANCE
from .room_generation import generate_room
from .structure_structuring import structure_rooms
from .biome_determination import determine_biome


def generate_structure(noise_offset, chunk_x, chunk_y, chunks, checked):
    if (chunk_x, chunk_y) not in checked:
        checked.add((chunk_x, chunk_y))
        structure_value = random()
        structure = False
        biome = determine_biome(16 * chunk_x, 16 * chunk_y, noise_offset)
        for noise_structure in NOISE_STRUCTURES.get(biome, ()):
            if noise_structure[0][0] < structure_value < noise_structure[0][1]:
                structure_type = noise_structure[1]
                structure = True
                break
        if structure:
            dungeon = structure_rooms(structure_type, (chunk_x, chunk_y))
            for dungeon_room in dungeon[0]:
                chunks[dungeon_room] = {}
                room = generate_room(structure_type, dungeon[0][dungeon_room], (1, 1), (chunk_x, chunk_y), dungeon[0][dungeon_room][0] == dungeon[0][dungeon_room][1])
                for chunk in room:
                    chunks[chunk] = room[chunk]
            if dungeon[2] not in chunks:
                chunks[dungeon[2]] = {}
            chunks[dungeon[2]][7, 1] = STRUCTURE_ENTRANCE[structure_type]
    return chunks, checked


import pygame as pg

from ...info import UI_SCALE, SCREEN_SIZE, BIG_UI_FONT, UI_FONT, SLOT_SIZE, TILE_UI_SIZE, FLOOR_SIZE, HALF_SCREEN_SIZE, RECIPES, FLOOR

def render_craft(window, machine_ui, images, recipe_number):
    current_recipes = RECIPES[machine_ui]
    window.blit(pg.transform.scale(images["big_inventory_slot_2"], (96 * UI_SCALE, 96 * UI_SCALE)), (HALF_SCREEN_SIZE - 128 * UI_SCALE, SCREEN_SIZE[1] - 144 * UI_SCALE))
    if current_recipes[recipe_number][0][0] not in FLOOR:
        window.blit(pg.transform.scale(images[current_recipes[recipe_number][0][0]], (48 * UI_SCALE, 72 * UI_SCALE)), (HALF_SCREEN_SIZE - 104 * UI_SCALE, SCREEN_SIZE[1] - 132 * UI_SCALE))
    else:
        window.blit(pg.transform.scale(images[current_recipes[recipe_number][0][0]], (48 * UI_SCALE, 48 * UI_SCALE)), (HALF_SCREEN_SIZE - 104 * UI_SCALE, SCREEN_SIZE[1] - 108 * UI_SCALE))
    window.blit(BIG_UI_FONT.render(str(current_recipes[recipe_number][0][1]), False, (19, 17, 18)), (HALF_SCREEN_SIZE - 112 * UI_SCALE, SCREEN_SIZE[1] - 80 * UI_SCALE),)
    for inputs in range(0, len(current_recipes[recipe_number][1])):
        position = (HALF_SCREEN_SIZE + ((40 * (inputs % 4) - 32)) * UI_SCALE, SCREEN_SIZE[1] + (32 * (inputs // 4) - 144) * UI_SCALE)
        window.blit(pg.transform.scale(images["inventory_slot"], SLOT_SIZE), position)
        if current_recipes[recipe_number][1][inputs][0] not in FLOOR:
            window.blit(pg.transform.scale(images[current_recipes[recipe_number][1][inputs][0]], TILE_UI_SIZE), (position[0] + 8 * UI_SCALE, position[1] + 4 * UI_SCALE),)
        else:
            window.blit(pg.transform.scale(images[current_recipes[recipe_number][1][inputs][0]], FLOOR_SIZE), (position[0] + 8 * UI_SCALE, position[1] + 12 * UI_SCALE),)
        window.blit(UI_FONT.render(str(current_recipes[recipe_number][1][inputs][1]), False, (19, 17, 18)), (position[0] + 8 * UI_SCALE, position[1] + 32 * UI_SCALE))
    return window


from ...info import INVENTORY_SIZE, HEALTH
from .damage_calculation import calculate_damage

def break_floor(mining_tile, inventory, inventory_number, player_tile):
    delete_mining_tile = False
    if "health" not in mining_tile:
        mining_tile["health"] = HEALTH[mining_tile["floor"]]
    mining_tile["health"] -= calculate_damage(mining_tile["floor"], inventory, inventory_number)
    broke = False
    if mining_tile["health"] <= 0:
        if mining_tile["floor"] in inventory:
            if inventory[mining_tile["floor"]] < INVENTORY_SIZE[1]:
                broke = True
        elif len(inventory) < INVENTORY_SIZE[0]:
            broke = True
    if broke:
        player_tile["inventory"][mining_tile["floor"]] = player_tile["inventory"].get(mining_tile["floor"], 0) + 1
        delete_mining_tile = True
    return delete_mining_tile, inventory, mining_tile


def move_player(key, controls, velocity, location):
    if key[controls[24]]:
        speed = 0.07
    else:
        speed = 0.1
    if key[controls[0]]:
        velocity[1] -= speed / (1 + abs(velocity[1]))
    if key[controls[1]]:
        velocity[0] -= speed / (1 + abs(velocity[1]))
    if key[controls[2]]:
        velocity[1] += speed / (1 + abs(velocity[1]))
    if key[controls[3]]:
        velocity[0] += speed / (1 + abs(velocity[1]))
    location["real"][2] += velocity[0] / 2
    location["real"][3] += velocity[1] / 2
    velocity[0] *= 0.65
    velocity[1] *= 0.65
    location["real"] = [
        location["real"][0] + location["real"][2] // 16,
        location["real"][1] + location["real"][3] // 16,
        location["real"][2] % 16,
        location["real"][3] % 16,
    ]
    location["tile"] = [
        int(location["real"][0]),
        int(location["real"][1]),
        int(location["real"][2]),
        int(location["real"][3]),
    ]
    return (location, velocity)



from ast import literal_eval
from os import listdir, path, remove

from ...render.menu_rendering import SAVES_FOLDER

def save_loading(state, chunks):
    saves = [f[:-4] for f in listdir(SAVES_FOLDER) if f.endswith(".txt")]
    index = (state.position[1] // 50) - 2
    if index < len(saves):
        state.save_file_name = saves[index]
        if state.position[0] >= 120:
            state.menu_placement = "main_game"
            with open(path.join(SAVES_FOLDER, state.save_file_name + ".txt"), "r", encoding="utf-8") as file:
                file_content = file.read().split(";")
            chunks = literal_eval(file_content[0])
            state.location["tile"] = literal_eval(file_content[1])
            state.tick = int(float(file_content[2]))
            state.location["real"] = list(state.location["tile"])
            state.noise_offset = literal_eval(file_content[3])
            state.world_type = int(file_content[4])
        elif state.position[0] <= 90:
            file_path = path.join(SAVES_FOLDER, state.save_file_name + ".txt")
            if path.exists(file_path):
                remove(file_path)
    return chunks


import pygame as pg

pg.display.init()
pg.font.init()

SCREEN_SIZE = (pg.display.Info().current_w, pg.display.Info().current_h)
TILE_SIZE = 64
HALF_SIZE = TILE_SIZE // 2
CHUNK_SIZE = 16 * TILE_SIZE
FPS = 60
DAY_LENGTH = 60 * 24 * FPS
UI_SCALE = 2
UI_FONT = pg.font.SysFont("Lucida Console", 10 * UI_SCALE)
BIG_UI_FONT = pg.font.SysFont("Lucida Console", 20 * UI_SCALE)
SLOT_SIZE = (32 * UI_SCALE, 32 * UI_SCALE)
TILE_UI_SIZE = (16 * UI_SCALE, 24 * UI_SCALE)
FLOOR_SIZE = (16 * UI_SCALE, 16 * UI_SCALE)
HALF_SCREEN_SIZE = SCREEN_SIZE[0] // 2
INVENTORY_SIZE = (12, 64)


TOOL_REQUIRED = {
    "big rock": "pickaxe",
    "bluebell": "shovel",
    "bonsai pot": "axe",
    "brick": "pickaxe",
    "campfire": "axe",
    "carrot": "shovel",
    "carroot": "shovel",
    "clay": "shovel",
    "coal": "pickaxe",
    "coal ore": "pickaxe",
    "composter": "axe",
    "copper brick": "pickaxe",
    "copper brick floor": "pickaxe",
    "dirt": "shovel",
    "flint": "pickaxe",
    "ice": "pickaxe",
    "log": "axe",
    "log floor": "axe",
    "manual press": "axe",
    "masonry bench": "pickaxe",
    "mushroom": "axe",
    "mushroom block": "hammer",
    "mushroom door": "hammer",
    "mushroom door open": "hammer",
    "mushroom floor": "hammer",
    "mushroom shaper": "hammer",
    "pebble": "pickaxe",
    "player": "sword",
    "rabbit adult": "sword",
    "rabbit child": "sword",
    "rock": "pickaxe",
    "sand": "shovel",
    "sapling": "shovel",
    "sawbench": "axe",
    "small crate": "axe",
    "spore": "shovel",
    "stick": "axe",
    "stone": "pickaxe",
    "stone brick": "pickaxe",
    "tree": "axe",
    "treeling": "axe",
    "wood": "axe",
    "wood door": "axe",
    "wood door open": "axe",
    "wood floor": "axe",
    "wood pulp": "shovel",
}
TOOL_EFFICIENCY = {
    "flint": 1, 
    "rock": 2
}


from .render_info import FPS

FERTILIZER_SPAWN = (
    (0.02, "bluebell"),
    (0.1, "carrot"),
    (0.3, "spore"),
    (0.6, "sapling"),
)
FERTILIZER_EFFICIENCY = {
    "fertilizer": 20,
    "compost": 1,
}
FOOD = {
    "carrot": 2,
    "mushroom": 1,
    "mushroom stew": 6,
    "rabbit meat": 1,
    "roasted mushroom": 3,
    "roasted rabbit meat": 4,
}
FLOOR = (
    "dirt",
    "ice",
    "log floor",
    "mushroom door",
    "mushroom door open",
    "mushroom floor",
    "pebble",
    "void",
    "water",
    "wood door",
    "wood door open",
    "wood floor",
)
FLOOR_TYPE = {
    "dirt": "soil",
    "ice": "block",
    "mushroom door": "door",
    "mushroom door open": "open",
    "void": "block",
    "water": "fluid",
    "wood door": "door",
    "wood door open": "open",
}
GROW_CHANCE = {
    "carrot": 160,
    "rabbit child": 200,
    "sapling": 80,
    "spore": 120,
    "treeling": 160,
}
GROW_TILES = {
    "carrot": {"kind": "carroot", "inventory": {"carrot": 2}},
    "rabbit child": {"kind": "rabbit adult", "inventory": {"rabbit fur": 1, "rabbit meat": 2}},
    "sapling": {"kind": "treeling", "inventory": {"sapling": 1, "log": 1}},
    "spore": {"kind": "mushroom", "inventory": {"mushroom": 1, "spore": 2}},
    "treeling": {"kind": "tree", "inventory": {"sapling": 2, "log": 2}},
}
MULTI_TILES = {
    "big rock": (2, 2),
    "manual press": (2, 1),
    "masonry bench": (2, 1),
    "obelisk": (1, 2),
    "sawbench": (2, 1),
    "wooden bed": (1, 2),
}
PROCESSING_TIME = {
    "bonsai pot": 40 * FPS,
    "composter": 2 * FPS,
}
STORAGE = {
    "small barrel": (1, 512),
    "small crate": (9, 48),
}
UNBREAK = (
    "glass lock",
    "left",
    "obelisk",
    "rabbit hole",
    "up",
    "void",
)


from json import dumps
from os import path

from ...render.menu_rendering import SAVES_FOLDER

def option(state, chunks):
    if state.menu_placement == "options_game":
        if 0 <= state.position[1] <= 50:
            state.menu_placement = "main_game"
        elif 100 <= state.position[1] <= 150:
            if state.save_file_name != "" and state.save_file_name.split("_")[0] != "autosave":
                state.menu_placement = "main_menu"
                with open(path.join(SAVES_FOLDER, state.save_file_name + ".txt"), "w", encoding="utf-8") as file:
                    file.write(f"{chunks};{state.location['tile']};{state.tick};{state.noise_offset};{state.world_type}")
                state.save_file_name = ""
                state.machine_ui = "game"
            else:
                state.menu_placement = "save_creation"
    elif state.menu_placement == "options_main":
        if 0 <= state.position[1] <= 50:
            state.menu_placement = "main_menu"
    if 200 <= state.position[1] <= 250:
        state.control_adjusted = -1
        state.menu_placement = "controls_options"


from ...info import ATTRIBUTES, DAY_LENGTH, FLOOR_TYPE
from ..left_click import recipe, place, open_storage, closed_storage, machine_storage, unlock, fertilize_grow, fertilize_spawn


def left_click(
    machine_ui: str,
    grid_position: list[int, int],
    chunks,
    inventory_number: int,
    health: int,
    position,
    recipe_number: int,
    location: dict[str],
    inventory: dict[str, int],
    machine_inventory: dict[str, int],
    tick: int,
):
    if machine_ui == "game":
        is_not_tile = (grid_position[1] not in chunks[grid_position[0]])
        if is_not_tile:
            is_kind = True
        else:
            is_kind = "kind" in chunks[grid_position[0]][grid_position[1]]
            current_tile = chunks[grid_position[0]][grid_position[1]]
        is_floor = not is_not_tile and not is_kind
        if is_floor and FLOOR_TYPE.get(current_tile["floor"]) == "door":
            chunks[grid_position[0]][grid_position[1]]["floor"] += " open"
        elif is_floor and FLOOR_TYPE.get(current_tile["floor"]) == "open":
            chunks[grid_position[0]][grid_position[1]]["floor"] = current_tile["floor"][:-5]
        elif is_floor and FLOOR_TYPE.get(current_tile["floor"]) == "soil" and inventory_number < len(inventory) and "fertilize" in ATTRIBUTES.get(list(inventory.keys())[inventory_number], ()):
            chunks = fertilize_spawn(chunks, inventory, inventory_number, grid_position)
        elif is_not_tile or not is_kind:
            chunks = place(inventory, inventory_number, is_not_tile, is_kind, health, grid_position, location, chunks)
        else:
            attributes = ATTRIBUTES.get(chunks[grid_position[0]][grid_position[1]]["kind"], ())
            if "open" in attributes:
                machine_ui = chunks[grid_position[0]][grid_position[1]]["kind"]
                location["opened"] = (grid_position[0], grid_position[1])
                machine_inventory = chunks[grid_position[0]][grid_position[1]].get("inventory", {})
            elif "sleep" in attributes:
                if 9 / 16 <= (tick / DAY_LENGTH) % 1 < 15 / 16:
                    tick = (tick // DAY_LENGTH + 9 / 16) * DAY_LENGTH
            elif "lock" in attributes:
                chunks = unlock(inventory, inventory_number, chunks, grid_position)
            elif "grow" in attributes and inventory_number < len(inventory):
                chunks = fertilize_grow(chunks, inventory, inventory_number, grid_position)
            elif "store" in attributes:
                chunks = closed_storage(chunks, grid_position, inventory, location, inventory_number)
    elif "machine" in ATTRIBUTES.get(machine_ui, ()):
        chunks = machine_storage(position, chunks, location, inventory, machine_ui)
    elif "store" in ATTRIBUTES.get(machine_ui, ()):
        chunks = open_storage(position, chunks, location, inventory, machine_ui)
    elif "craft" in ATTRIBUTES.get(machine_ui, ()):
        inventory = recipe(machine_ui, recipe_number, inventory)
    return machine_ui, chunks, location, machine_inventory, tick



from .game_state import GameState
from ..updates import update_menu, update_game, update_tiles


def update(state: GameState, chunks):
    if state.menu_placement != "main_game":
        chunks = update_menu(state, chunks)
    else:
        chunks = update_game(state, chunks)
        chunks = update_tiles(state, chunks)
    return chunks


from ...info import RESISTANCE, TOOL_EFFICIENCY, TOOL_REQUIRED

def calculate_damage(mining_type, inventory, inventory_number):
    damage = 1 - RESISTANCE.get(mining_type, 0)
    if len(inventory) > inventory_number:
        inventory_words = list(inventory.keys())[inventory_number].split()
        if len(inventory_words) == 2 and mining_type in TOOL_REQUIRED:
            if TOOL_REQUIRED[mining_type] == inventory_words[1]:
                damage += TOOL_EFFICIENCY[inventory_words[0]]
    return max(damage, 0)


import os
from PIL import Image, ImageOps
from random import randint

from ..info import ROOM_COLORS
from .loot_calculation import calculate_loot


def generate_room(structure, room, tile_offset, chunk_offset, rotate = True, vary = True):
    room_image = Image.open(os.path.normpath(os.path.join(__file__, "../../..", f"structures/{structure}/{room}.png"))).convert("RGB")
    room_chunks = {}
    if vary:
        variation = randint(0, 15)
        if variation % 2:
            room_image = ImageOps.mirror(room_image)
        if (variation // 2) % 2:
            room_image = ImageOps.flip(room_image)
        if rotate:
            room_image = room_image.rotate((variation // 4) * 90)
    for x in range(0, room_image.size[0]):
        for y in range(0, room_image.size[1]):
            if room_image.getpixel((x, y)) in ROOM_COLORS:
                tile_placement = (x + tile_offset[0]) % 16, (y + tile_offset[1]) % 16
                chunk_placement = (x + tile_offset[0]) // 16 + chunk_offset[0], (y + tile_offset[1]) // 16 + chunk_offset[1]
                if chunk_placement not in room_chunks:
                    room_chunks[chunk_placement] = {}
                room_chunks[chunk_placement][tile_placement] = {}
                tile = ROOM_COLORS[room_image.getpixel((x, y))]
                for index in tile:
                    room_chunks[chunk_placement][tile_placement][index] = tile[index]
                if "loot" in room_chunks[chunk_placement][tile_placement]:
                    room_chunks[chunk_placement][tile_placement] = calculate_loot(room_chunks[chunk_placement][tile_placement])
    return room_chunks


from ...info import INVENTORY_SIZE

def take_out(chunks, location, inventory, item):
    inventory_item = inventory.get(item[0], 0)
    if not (inventory_item == 0 and len(inventory) == INVENTORY_SIZE[0]):
        if inventory_item + item[1] <= INVENTORY_SIZE[1]:
            chunks[location["tile"][0], location["tile"][1]][location["tile"][2], location["tile"][3]]["inventory"][item[0]] = inventory_item + item[1]
            del chunks[location["opened"][0]][location["opened"][1]]["inventory"][item[0]]
        else:
            chunks[location["tile"][0], location["tile"][1]][location["tile"][2], location["tile"][3]]["inventory"][item[0]] = INVENTORY_SIZE[1]
            chunks[location["opened"][0]][location["opened"][1]]["inventory"][item[0]] = inventory_item + item[1] - INVENTORY_SIZE[1]
    if chunks[location["opened"][0]][location["opened"][1]]["inventory"] == {}:
        del chunks[location["opened"][0]][location["opened"][1]]["inventory"]
    return chunks


from random import random, choice
from ..info import STRUCTURE_SIZE, STRUCTURE_ROOM_SIZES, STRUCTURE_ROOMS


ADJACENT_ROOMS = ((0, -1), (0, 1), (-1, 0), (1, 0))

def structure_structure(dungeon_type, offset, dungeon=None, tile=None, distanse=1):
    if dungeon == None:
        dungeon = set()
    if tile == None:
        tile = offset
    if tile != (0, 0):
        dungeon.add(tile)
    for pos in ADJACENT_ROOMS:
        if random() < STRUCTURE_SIZE[dungeon_type]**distanse and (tile[0] + pos[0], tile[1] + pos[1]) not in dungeon:
            structure_structure(dungeon_type, offset, dungeon, (tile[0] + pos[0], tile[1] + pos[1]), distanse + 1)
    return dungeon

def add_hallways(hallways, room, adj_room):
    if isinstance(room, tuple) and isinstance(adj_room, tuple):
        hallways.setdefault(room, set()).add(adj_room)
        hallways.setdefault(adj_room, set()).add(room)
    return hallways

def structure_hallways(room, dungeon, hallways=None, visited=None):
    if hallways == None:
        hallways = {}
    if visited == None:
        visited = set()
    visited.add(room)
    for pos in ADJACENT_ROOMS:
        adj_room = (room[0] + pos[0], room[1] + pos[1])
        if adj_room in dungeon:
            if random() < 0.5:
                hallways = add_hallways(hallways, room, adj_room)
                if adj_room not in visited:
                    hallways = structure_hallways((room[0] + pos[0], room[1] + pos[1]), dungeon, hallways, visited=visited)
    return hallways

def ensure_hallways(dungeon, hallways, room, visited=None):
    if visited == None:
        visited = set()
    visited.add(room)
    adj_dungeon_rooms = []
    for pos in ADJACENT_ROOMS:
        adj_room = (room[0] + pos[0], room[1] + pos[1])
        if adj_room not in visited:
            if adj_room in hallways:
                hallways = add_hallways(hallways, room, adj_room)
                return hallways
            elif adj_room in dungeon:
                adj_dungeon_rooms.append(adj_room)
    if len(adj_dungeon_rooms):
        next_room = choice(adj_dungeon_rooms)
        hallways = add_hallways(hallways, room, next_room)
        hallways = ensure_hallways(dungeon, hallways, next_room, visited)
    return hallways

def structure_rooms(dungeon_type, offset):
    structure = structure_structure(dungeon_type, offset)
    dungeon = {}
    if len(structure) > 1:
        hallways = structure_hallways((0, 0), structure)
        for room in structure:
            if room not in hallways:
                hallways = ensure_hallways(structure, hallways, room)
    else:
        hallways = {}
    y = 0
    while (offset[0], y + offset[1]) in structure:
        y -= 1
    entrance = (offset[0], y + 1 + offset[1])
    for room in structure:
        if room not in dungeon:
            avaliable_sizes = []
            for size in STRUCTURE_ROOM_SIZES[dungeon_type]:
                can_place = True
                for x in range(0, size[0]):
                    for y in range(0, size[1]):
                        if (room[0] + x, room[1] + y) in dungeon:
                            can_place = False
                if can_place:
                    avaliable_sizes.append(size)
            size = choice(avaliable_sizes)
            for x in range(0, size[0]):
                for y in range(0, size[1]):
                    dungeon[room[0] + x, room[1] + y] = (-x, -y)
            dungeon[room] = size
    delete_rooms = set()
    for room in dungeon:
        if dungeon[room][0] > 0 < dungeon[room][1]:
            dungeon[room] = choice(STRUCTURE_ROOMS[dungeon_type][dungeon[room]])
        else:
            delete_rooms.add(room)
    for room in delete_rooms:
        del dungeon[room]
    return (dungeon, hallways, entrance)


import pygame as pg

from ...info import FLOOR, SCREEN_SIZE, UI_SCALE, SLOT_SIZE, TILE_UI_SIZE, UI_FONT, FLOOR_SIZE, HALF_SCREEN_SIZE, RECIPES

def render_machine(window, machine_ui, images, machine_inventory, recipe_number):
    machine_recipe = RECIPES[machine_ui][recipe_number]
    for item in range(0, len(machine_recipe[1])):
        window.blit(pg.transform.scale(images["inventory_slot"], SLOT_SIZE), (HALF_SCREEN_SIZE + (32 * (item % 7) - 112) * UI_SCALE, SCREEN_SIZE[1] + (32 * (item // 7) - 144) * UI_SCALE))
    window.blit(pg.transform.scale(images["inventory_slot_2"], SLOT_SIZE), (HALF_SCREEN_SIZE - 112 * UI_SCALE, SCREEN_SIZE[1]- 80 * UI_SCALE))
    for i in range(0, len(machine_recipe[1])):
        position = (HALF_SCREEN_SIZE + (32 * (i % 7) - 104) * UI_SCALE, SCREEN_SIZE[1] + (32 * (i // 7) - 140) * UI_SCALE)
        item = machine_recipe[1][i][0]
        if item not in FLOOR:
            window.blit(pg.transform.scale(images[item], TILE_UI_SIZE), position)
        else:
            window.blit(pg.transform.scale(images[item], FLOOR_SIZE), (position[0], position[1] + 8 * UI_SCALE))
        window.blit(UI_FONT.render(f"{machine_inventory.get(item, 0)}/{machine_recipe[1][i][1]}", False, (19, 17, 18)), (position[0] - 4 * UI_SCALE, position[1]))
    position = (HALF_SCREEN_SIZE - 104 * UI_SCALE, SCREEN_SIZE[1] - 76 * UI_SCALE)
    item = machine_recipe[0][0]
    if item not in FLOOR:
        window.blit(pg.transform.scale(images[item], TILE_UI_SIZE), position)
    else:
        window.blit(pg.transform.scale(images[item], FLOOR_SIZE), (position[0], position[1] + 8 * UI_SCALE))
    window.blit(UI_FONT.render(f"{machine_inventory.get(item, 0)}/{machine_recipe[0][1]}", False, (19, 17, 18)), (position[0] - 4 * UI_SCALE, position[1]))
    return window


import pygame as pg

from ...info import TILE_SIZE

def render_border(chunk, x, y, chunks, placement, zoom, window, current_tile):
    total_x = chunk[0] * 16 + x
    total_y = chunk[1] * 16 + y
    for dx, dy in [(-1, 0), (1, 0), (0, -1), (0, 1), (-1, -1), (1, -1), (-1, 1), (1, 1)]:
        adjacent_chunk = ((total_x + dx) // 16, (total_y + dy) // 16)
        adjacent_tile = ((total_x + dx) % 16, (total_y + dy) % 16)
        if adjacent_tile not in chunks[adjacent_chunk] or "floor" not in chunks[adjacent_chunk][adjacent_tile] or chunks[adjacent_chunk][adjacent_tile]["floor"] != current_tile["floor"]:
            if dx != 0 and dy != 0:
                rect = pg.Rect(placement[0] + ((TILE_SIZE - 2) * zoom if dx == 1 else 0), placement[1] + ((TILE_SIZE - 2) * zoom if dy == 1 else 0), zoom * 4, zoom * 4)
            elif dx != 0:
                rect = pg.Rect(placement[0] + dx * (TILE_SIZE - 2) * zoom if dx == 1 else placement[0], placement[1], zoom * 4, (TILE_SIZE + 2) * zoom)
            else:
                rect = pg.Rect(placement[0], placement[1] + dy * (TILE_SIZE - 2) * zoom if dy == 1 else placement[1], (TILE_SIZE + 2) * zoom, zoom * 4)
            pg.draw.rect(window, (19, 17, 18), rect)


import pygame as pg

from ...info import TILE_SIZE, HALF_SIZE, CHUNK_SIZE

def render_hand(inventory, inventory_number, camera, location, zoom, window, images):
    if len(inventory) > inventory_number:
        placement = (camera[0] + (location["tile"][2] * TILE_SIZE + location["tile"][0] * CHUNK_SIZE - 4) * zoom, camera[1] + (location["tile"][3] * TILE_SIZE + location["tile"][1] * CHUNK_SIZE - 8) * zoom)
        hand_image = pg.transform.scale(images[list(inventory.keys())[inventory_number]], (HALF_SIZE * zoom, 3 * TILE_SIZE // 4 * zoom))
        window.blit(hand_image, placement)
    return window


import pygame as pg

from ...info import SCREEN_SIZE

def update_keys(menu_placement, save_file_name, controls, control_adjusted, chunks, scroll):
    keys = pg.key.get_pressed()
    if menu_placement == "save_creation":
        for letter in range(48, 123):
            if keys[letter]:
                save_file_name += chr(letter)
        if keys[pg.K_SPACE]:
            save_file_name += " "
        elif keys[pg.K_BACKSPACE]:
            save_file_name = save_file_name[:-1]

    elif menu_placement == "controls_options":
        if control_adjusted > -1:
            for key_code in range(len(keys)):
                if keys[key_code]:
                    controls[control_adjusted] = key_code
                    control_adjusted = -1
        else:
            if keys[controls[21]]:
                menu_placement = "options_game" if chunks != {} else "options_main"
            elif keys[controls[22]]:
                if scroll < len(controls) - SCREEN_SIZE[1] // 50 - 1:
                    scroll += 1
            elif keys[controls[23]]:
                if scroll > 0:
                    scroll -= 1
    return save_file_name, controls, menu_placement, scroll, control_adjusted


from .attribute_info import ATTRIBUTES
from .health_info import HEALTH
from .recipes_info import RECIPES
from .render_info import SCREEN_SIZE, SLOT_SIZE, UI_SCALE, HALF_SCREEN_SIZE, HALF_SIZE, TILE_SIZE, CHUNK_SIZE, FLOOR_SIZE, TILE_UI_SIZE, INVENTORY_SIZE, FPS, UI_FONT, BIG_UI_FONT, DAY_LENGTH
from .resistance_info import RESISTANCE
from .tile_info import MULTI_TILES, STORAGE, FOOD, FLOOR, FLOOR_TYPE, UNBREAK, GROW_CHANCE, GROW_TILES, FERTILIZER_SPAWN, FERTILIZER_EFFICIENCY
from .tool_info import TOOL_EFFICIENCY, TOOL_REQUIRED
from .world_gen_info import NOISE_TILES, ATTRIBUTE_CARE, BIOMES, WORLD_TYPES
from .structure_gen_info import NOISE_STRUCTURES, STRUCTURE_SIZE, STRUCTURE_ROOM_SIZES, STRUCTURE_ROOMS, STRUCTURE_ENTRANCE, ROOM_COLORS, LOOT_TABLES
from .value_info import VALUES, MACHINES


import pygame as pg

from ...info import FLOOR, STORAGE, SCREEN_SIZE, UI_SCALE, SLOT_SIZE, TILE_UI_SIZE, UI_FONT, FLOOR_SIZE, HALF_SCREEN_SIZE

def render_store(window, machine_ui, images, machine_inventory):
    for item in range(0, STORAGE[machine_ui][0]):
        window.blit(pg.transform.scale(images["inventory_slot"], SLOT_SIZE), (HALF_SCREEN_SIZE + (32 * (item % 7) - 112) * UI_SCALE, SCREEN_SIZE[1] + (32 * (item // 7) - 144) * UI_SCALE))
    t = 0
    for item in machine_inventory:
        position = (HALF_SCREEN_SIZE + (32 * (t % 7) - 104) * UI_SCALE, SCREEN_SIZE[1] + (32 * (t // 7) - 140) * UI_SCALE)
        if item not in FLOOR:
            window.blit(pg.transform.scale(images[item], TILE_UI_SIZE), position)
        else:
            window.blit(pg.transform.scale(images[item], FLOOR_SIZE), (position[0], position[1] + 8 * UI_SCALE))
        window.blit(UI_FONT.render(str(machine_inventory[item]), False, (19, 17, 18)), position)
        t += 1
    return window


from .craft import recipe
from .placement import place
from .store import open_storage, closed_storage
from .machine import machine_storage
from .open_lock import unlock
from .fertilize import fertilize_spawn, fertilize_grow


import pygame as pg

from .player_move import move_player
from ...world_generation.world_generation import generate_chunk
from ...world_generation.structure_generation import generate_structure
from ...other_systems.game_state import GameState
from .mouse_update import button_press
from os import path
from ...info import DAY_LENGTH, INVENTORY_SIZE, FLOOR_TYPE, HEALTH
from ...render.menu_rendering import SAVES_FOLDER

def update_game(state: GameState, chunks):
    state.health = chunks[state.location["tile"][0], state.location["tile"][1]][state.location["tile"][2], state.location["tile"][3]].get("health", 20)
    state.location["old"] = list(state.location["tile"])
    state.inventory = chunks[state.location["tile"][0], state.location["tile"][1]][state.location["tile"][2], state.location["tile"][3]].get("inventory", {})
    key = pg.key.get_pressed()
    state.location, state.velocity = move_player(key, state.controls, state.velocity, state.location)

    for x in range(-4, 5):
        for y in range(-4, 5):
            generate_chunk(state.world_type, state.location["tile"][0] + x, state.location["tile"][1] + y, chunks, state.noise_offset)
    if state.world_type == 0:
        for x in range(-20, 20):
            for y in range(-20, 20):
                generate_structure(state.noise_offset, state.location["tile"][0] + x, state.location["tile"][1] + y, chunks, state.checked)

    tile_chunk_coords = (state.location["tile"][0], state.location["tile"][1])
    tile_coords = (state.location["tile"][2], state.location["tile"][3])
    old_chunk_coords = (state.location["old"][0], state.location["old"][1])
    old_tile_coords = (state.location["old"][2], state.location["old"][3])

    chunk = chunks[tile_chunk_coords]
    old_chunk = chunks[old_chunk_coords]
    old_tile = old_chunk[old_tile_coords]

    if tile_coords not in chunk:
        chunk[tile_coords] = {"kind": "player", "inventory": state.inventory, "health": state.health}
    elif "kind" not in chunk[tile_coords] and "door" != FLOOR_TYPE.get(chunk[tile_coords]["floor"]) != "fluid":
        exist_tile = chunk[tile_coords]
        chunk[tile_coords] = {"kind": "player", "inventory": state.inventory, "health": state.health, "floor": exist_tile["floor"]}
    elif chunk[tile_coords].get("kind") != "player":
        state.location["real"] = list(state.location["old"])
        state.location["tile"] = list(state.location["old"])
        state.velocity = [0, 0]

    if state.location["old"] != state.location["tile"]:
        if "floor" in old_tile:
            old_chunk[old_tile_coords] = {"floor": old_tile["floor"], "health": HEALTH.get(old_tile["floor"], 1)}
        else:
            del old_chunk[old_tile_coords]

    for event in pg.event.get():
        if event.type == pg.QUIT:
            state.run = False
        elif event.type == pg.MOUSEBUTTONDOWN:
            chunks, state.location, state.machine_ui, state.machine_inventory, state.tick, state.inventory_number = button_press(
                event.button, state.position, state.zoom, chunks, state.location, state.machine_ui, state.inventory, state.health,
                state.machine_inventory, state.tick, state.inventory_number, chunks[state.location["opened"][0]][state.location["opened"][1]].get("recipe", 0), state.camera)
        elif event.type == pg.KEYDOWN:
            keys = pg.key.get_pressed()
            if keys[state.controls[4]]:
                if state.machine_ui == "game":
                    state.machine_ui = "player"
                    state.location["opened"] = ((state.location["tile"][0], state.location["tile"][1]), (state.location["tile"][2], state.location["tile"][3]))
                else:
                    state.machine_ui = "game"
                    state.location["opened"] = ((0, 0), (0, 0))
            elif keys[state.controls[5]] or keys[state.controls[6]]:
                state.target_zoom += (keys[state.controls[5]] - keys[state.controls[6]]) / 4
                state.target_zoom = min(max(state.target_zoom, 0.5), 2)
            elif keys[state.controls[21]]:
                state.menu_placement = "options_game"
            elif keys[state.controls[0]] or keys[state.controls[1]] or keys[state.controls[2]] or keys[state.controls[3]]:
                state.machine_ui = "game"
                state.location["opened"] = ((0, 0), (0, 0))
            elif keys[state.controls[19]]:
                state.inventory_number = (state.inventory_number + 1) % INVENTORY_SIZE[0]
            elif keys[state.controls[20]]:
                state.inventory_number = (state.inventory_number - 1) % INVENTORY_SIZE[0]
            for i in range(7, 19):
                if keys[state.controls[i]]:
                    state.inventory_number = i - 7

    state.tick += 1
    if state.tick % (DAY_LENGTH // 4) == 0:
        with open(path.join(SAVES_FOLDER, f"autosave_{(state.tick // (DAY_LENGTH // 4)) % 4}.txt"), "w", encoding="utf-8") as file:
                file.write(f"{chunks};{state.location['tile']};{state.tick};{state.noise_offset}")
    state.zoom = 0.05 * state.target_zoom + 0.95 * state.zoom
    return chunks


from noise import pnoise2
from random import uniform

from ..info import MULTI_TILES, NOISE_TILES, ATTRIBUTE_CARE, ATTRIBUTES
from .biome_determination import determine_biome

def generate_chunk(
    world_type: int,
    chunk_x: int,
    chunk_y: int,
    chunks: dict[tuple[int, int], dict[tuple[int, int], tuple[dict, dict]]],
    noise_offset: tuple[float, float] = None
):
    if noise_offset == None:
        noise_offset = (uniform(-10000, 10000), uniform(-10000, 10000))
    if (chunk_x, chunk_y) not in chunks:
        chunks[chunk_x, chunk_y] = {}
        tile = chunks[chunk_x, chunk_y]
        if world_type == 0:
            for tile_x in range(0, 16):
                for tile_y in range(0, 16):
                    tile_pos = (tile_x, tile_y)
                    if tile_pos not in tile:
                        world_x = chunk_x * 16 + tile_x + noise_offset[0]
                        world_y = chunk_y * 16 + tile_y + noise_offset[1]
                        biome = determine_biome(world_x, world_y, noise_offset)
                        elevation_value = pnoise2(world_x / 10 + noise_offset[0], world_y / 10 + noise_offset[1], 3, 0.5, 2)
                        moisture_value = pnoise2(world_x / 30 + noise_offset[0], world_y / 30 + noise_offset[1], 3, 0.5, 2)
                        for noise_tile in NOISE_TILES[biome]:
                            if noise_tile[0][0] < elevation_value < noise_tile[0][1] and noise_tile[1][0] < moisture_value < noise_tile[1][1]:
                                tile[tile_pos] = noise_tile[2]
                                break
                        if tile_pos in tile:
                            tile_size = MULTI_TILES.get(tile[tile_pos].get("kind", None), (1, 1))
                            new_tile_x = tile_x - tile_size[0] + 1
                            new_tile_y = tile_y - tile_size[1] + 1
                            can_place = True
                            if tile_x - tile_size[0] + 1 < 0 or tile_y - tile_size[1] + 1 < 0:
                                can_place = False
                            for x in range(0, tile_size[0]):
                                for y in range(0, tile_size[1]):
                                    test_tile = (new_tile_x + x, new_tile_y + y)
                                    if test_tile in tile:
                                        if len(tuple(ATTRIBUTE_CARE & ATTRIBUTES.get(tile[test_tile].get("kind", None), set()))) or "structure" in tile[test_tile]:
                                            can_place = False
                            if can_place:
                                tile[new_tile_x, new_tile_y] = tile[tile_pos]
                                for x in range(0, tile_size[0]):
                                    if x != 0:
                                        tile[new_tile_x + x, new_tile_y] = {"kind": "left"}
                                    for y in range(1, tile_size[1]):
                                        tile[new_tile_x + x, new_tile_y + y] = {"kind": "up"}
                            else:
                                del tile[tile_pos]
    return noise_offset


import pygame as pg

from ..menu_update import update_keys, update_mouse

def update_menu(state, chunks):
    for event in pg.event.get():
        if event.type == pg.QUIT:
            state.run = False
        elif event.type == pg.MOUSEBUTTONDOWN:
            chunks = update_mouse(state, event, chunks)
        elif event.type == pg.KEYDOWN:
            state.save_file_name, state.controls, state.menu_placement, state.scroll, state.control_adjusted = update_keys(state.menu_placement, state.save_file_name, state.controls, state.control_adjusted, chunks, state.scroll)
    return chunks